# -*- coding: utf-8 -*-
from collections import OrderedDict
from pickle import dumps, loads
from unittest.mock import Mock

from pytest import raises

from pyee import EventEmitter


class PyeeTestException(Exception):
    pass


def test_emit_sync():
    """Basic synchronous emission works"""

    call_me = Mock()
    ee = EventEmitter()

    @ee.on("event")
    def event_handler(data, **kwargs):
        call_me()
        assert data == "emitter is emitted!"

    assert ee.event_names() == {"event"}

    # Making sure data is passed propers
    ee.emit("event", "emitter is emitted!", error=False)

    call_me.assert_called_once()


def test_emit_error():
    """Errors raise with no event handler, otherwise emit on handler"""

    call_me = Mock()
    ee = EventEmitter()

    test_exception = PyeeTestException("lololol")

    with raises(PyeeTestException):
        ee.emit("error", test_exception)

    @ee.on("error")
    def on_error(exc):
        call_me()

    assert ee.event_names() == {"error"}

    # No longer raises and error instead return True indicating handled
    assert ee.emit("error", test_exception) is True
    call_me.assert_called_once()


def test_emit_return():
    """Emit returns True when handlers are registered on an event, and false
    otherwise.
    """

    call_me = Mock()
    ee = EventEmitter()

    assert ee.event_names() == set()

    # make sure emitting without a callback returns False
    assert not ee.emit("data")

    # add a callback
    ee.on("data")(call_me)

    # should return True now
    assert ee.emit("data")


def test_new_listener_event():
    """The 'new_listener' event fires whenever a new listener is added."""

    call_me = Mock()
    ee = EventEmitter()

    ee.on("new_listener", call_me)

    # Should fire new_listener event
    @ee.on("event")
    def event_handler(data):
        pass

    assert ee.event_names() == {"new_listener", "event"}

    call_me.assert_called_once_with("event", event_handler)


def test_listener_removal():
    """Removing listeners removes the correct listener from an event."""

    ee = EventEmitter()

    # Some functions to pass to the EE
    def first():
        return 1

    ee.on("event", first)

    @ee.on("event")
    def second():
        return 2

    @ee.on("event")
    def third():
        return 3

    def fourth():
        return 4

    ee.on("event", fourth)

    assert ee.event_names() == {"event"}

    assert ee._events["event"] == OrderedDict(
        [(first, first), (second, second), (third, third), (fourth, fourth)]
    )

    ee.remove_listener("event", second)

    assert ee._events["event"] == OrderedDict(
        [(first, first), (third, third), (fourth, fourth)]
    )

    ee.remove_listener("event", first)
    assert ee._events["event"] == OrderedDict([(third, third), (fourth, fourth)])

    ee.remove_all_listeners("event")
    assert "event" not in ee._events["event"]


def test_listener_removal_on_emit():
    """Test that a listener removed during an emit is called inside the current
    emit cycle.
    """

    call_me = Mock()
    ee = EventEmitter()

    def should_remove():
        ee.remove_listener("remove", call_me)

    ee.on("remove", should_remove)
    ee.on("remove", call_me)

    assert ee.event_names() == {"remove"}

    ee.emit("remove")

    call_me.assert_called_once()

    call_me.reset_mock()

    # Also test with the listeners added in the opposite order
    ee = EventEmitter()
    ee.on("remove", call_me)
    ee.on("remove", should_remove)

    assert ee.event_names() == {"remove"}

    ee.emit("remove")

    call_me.assert_called_once()


def test_once():
    """Test that `once()` method works propers."""

    # very similar to "test_emit" but also makes sure that the event
    # gets removed afterwards

    call_me = Mock()
    ee = EventEmitter()

    def once_handler(data):
        assert data == "emitter is emitted!"
        call_me()

    # Tests to make sure that after event is emitted that it's gone.
    ee.once("event", once_handler)

    assert ee.event_names() == {"event"}

    ee.emit("event", "emitter is emitted!")

    call_me.assert_called_once()

    assert ee.event_names() == set()

    assert "event" not in ee._events


def test_once_removal():
    """Removal of once functions works"""

    ee = EventEmitter()

    def once_handler(data):
        pass

    handle = ee.once("event", once_handler)

    assert handle == once_handler
    assert ee.event_names() == {"event"}

    ee.remove_listener("event", handle)

    assert "event" not in ee._events
    assert ee.event_names() == set()


def test_listeners():
    """`listeners()` returns a copied list of listeners."""

    call_me = Mock()
    ee = EventEmitter()

    @ee.on("event")
    def event_handler():
        pass

    @ee.once("event")
    def once_handler():
        pass

    listeners = ee.listeners("event")

    assert listeners[0] == event_handler
    assert listeners[1] == once_handler

    # listeners is a copy, you can't mutate the innards this way
    listeners[0] = call_me

    ee.emit("event")

    call_me.assert_not_called()


def test_listeners_does_work_with_unknown_listeners():
    """`listeners()` should not throw."""
    ee = EventEmitter()
    listeners = ee.listeners("event")
    assert listeners == []


def test_properties_preserved():
    """Test that the properties of decorated functions are preserved."""

    call_me = Mock()
    call_me_also = Mock()
    ee = EventEmitter()

    @ee.on("always")
    def always_event_handler():
        """An event handler."""
        call_me()

    @ee.once("once")
    def once_event_handler():
        """Another event handler."""
        call_me_also()

    assert always_event_handler.__doc__ == "An event handler."
    assert once_event_handler.__doc__ == "Another event handler."

    always_event_handler()
    call_me.assert_called_once()

    once_event_handler()
    call_me_also.assert_called_once()

    call_me_also.reset_mock()

    # Calling the event handler directly doesn't clear the handler
    ee.emit("once")
    call_me_also.assert_called_once()


def test_if_serializable():
    """`pickle`ing should not throw."""
    ee = EventEmitter()
    ee_copy = loads(dumps(ee))
    assert ee._lock
    assert ee_copy._lock



# Configuration file for the Sphinx documentation builder.

# pyee uses mkdocs for its primary documentation. However, it uses sphinx to
# generate a man page.

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import toml

with open("../pyproject.toml", "r") as f:
    pyproject_toml = toml.load(f)

project = "pyee"
copyright = "2023, Josh Holbrook"
author = "Josh Holbrook"
release = f'v{pyproject_toml["project"]["version"]}'

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "index.md"]

root_doc = "man"



# -*- coding: utf-8 -*-

import asyncio
from asyncio import Future, get_running_loop, sleep, wait_for
from typing import NoReturn

import pytest
import pytest_asyncio.plugin  # noqa

try:
    from asyncio.exceptions import TimeoutError  # type: ignore
except ImportError:
    from concurrent.futures import TimeoutError  # type: ignore

from pyee.asyncio import AsyncIOEventEmitter


class PyeeTestError(Exception):
    pass


@pytest.mark.asyncio
async def test_emit() -> None:
    """Test that AsyncIOEventEmitter can handle wrapping
    coroutines
    """

    ee = AsyncIOEventEmitter(loop=get_running_loop())

    should_call: Future[bool] = Future(loop=get_running_loop())

    @ee.on("event")
    async def event_handler() -> None:
        should_call.set_result(True)

    ee.emit("event")

    result = await wait_for(should_call, 0.1)

    assert result is True

    await asyncio.sleep(0)
    assert ee.complete


@pytest.mark.asyncio
async def test_once_emit() -> None:
    """Test that AsyncIOEventEmitter also wrap coroutines when
    using once
    """

    ee = AsyncIOEventEmitter(loop=get_running_loop())

    should_call: Future[bool] = Future(loop=get_running_loop())

    @ee.once("event")
    async def event_handler():
        should_call.set_result(True)

    ee.emit("event")

    result = await wait_for(should_call, 0.1)

    assert result is True


@pytest.mark.asyncio
async def test_error() -> None:
    """Test that AsyncIOEventEmitter can handle errors when
    wrapping coroutines
    """
    ee = AsyncIOEventEmitter(loop=get_running_loop())

    should_call: Future[bool] = Future(loop=get_running_loop())

    @ee.on("event")
    async def event_handler() -> NoReturn:
        raise PyeeTestError()

    @ee.on("error")
    def handle_error(exc):
        should_call.set_result(exc)

    ee.emit("event")

    result = await wait_for(should_call, 0.1)

    assert isinstance(result, PyeeTestError)


@pytest.mark.asyncio
async def test_future_canceled() -> None:
    """Test that AsyncIOEventEmitter can handle canceled Futures"""

    cancel_me: Future[bool] = Future(loop=get_running_loop())
    should_not_call: Future[None] = Future(loop=get_running_loop())

    ee = AsyncIOEventEmitter(loop=get_running_loop())

    @ee.on("event")
    async def event_handler() -> None:
        cancel_me.cancel()

    @ee.on("error")
    def handle_error(exc) -> None:
        should_not_call.set_result(None)

    ee.emit("event")

    try:
        await wait_for(should_not_call, 0.1)
    except TimeoutError:
        pass
    else:
        raise PyeeTestError()


@pytest.mark.asyncio
async def test_event_emitter_canceled() -> None:
    """Test that all running handlers in AsyncIOEventEmitter can be canceled"""

    ee = AsyncIOEventEmitter(loop=get_running_loop())

    should_not_call: Future[bool] = Future(loop=get_running_loop())

    @ee.on("event")
    async def event_handler():
        await sleep(1)
        should_not_call.set_result(True)

    ee.emit("event")

    await sleep(0)

    # event_handler should still be running
    assert not ee.complete

    # cancel all pending tasks, including event_handler
    ee.cancel()

    await sleep(0)

    # event_handler should be canceled
    assert ee.complete


@pytest.mark.asyncio
async def test_wait_for_complete() -> None:
    """Test waiting for all pending tasks in an AsyncIOEventEmitter to
    complete
    """

    ee = AsyncIOEventEmitter(loop=get_running_loop())

    @ee.on("event")
    async def event_handler():
        await sleep(0.1)

    ee.emit("event")

    await sleep(0)

    # event_handler should still be running
    assert not ee.complete

    # wait for event_handler to complete execution
    await ee.wait_for_complete()

    # event_handler should have finished execution
    assert ee.complete


@pytest.mark.asyncio
async def test_sync_error() -> None:
    """Test that regular functions have the same error handling as coroutines"""
    ee = AsyncIOEventEmitter(loop=get_running_loop())

    should_call: Future[Exception] = Future(loop=get_running_loop())

    @ee.on("event")
    def sync_handler():
        raise PyeeTestError()

    @ee.on("error")
    def handle_error(exc):
        should_call.set_result(exc)

    ee.emit("event")

    result = await wait_for(should_call, 0.1)

    assert isinstance(result, PyeeTestError)



# -*- coding: utf-8 -*-

from unittest.mock import call, Mock

import pytest

from pyee import EventEmitter
from pyee.uplift import unwrap, uplift


class UpliftedEventEmitter(EventEmitter):
    pass


def test_uplift_emit():
    call_me = Mock()

    base_ee = EventEmitter()

    @base_ee.on("base_event")
    def base_handler():
        call_me("base event on base emitter")

    @base_ee.on("shared_event")
    def shared_base_handler():
        call_me("shared event on base emitter")

    uplifted_ee = uplift(UpliftedEventEmitter, base_ee)

    assert isinstance(uplifted_ee, UpliftedEventEmitter), "Returns an uplifted emitter"

    @uplifted_ee.on("uplifted_event")
    def uplifted_handler():
        call_me("uplifted event on uplifted emitter")

    @uplifted_ee.on("shared_event")
    def shared_uplifted_handler():
        call_me("shared event on uplifted emitter")

    # Events on uplifted proxy correctly
    assert uplifted_ee.emit("base_event")
    assert uplifted_ee.emit("shared_event")
    assert uplifted_ee.emit("uplifted_event")

    call_me.assert_has_calls(
        [
            call("base event on base emitter"),
            call("shared event on uplifted emitter"),
            call("shared event on base emitter"),
            call("uplifted event on uplifted emitter"),
        ]
    )

    call_me.reset_mock()

    # Events on underlying proxy correctly
    assert base_ee.emit("base_event")
    assert base_ee.emit("shared_event")
    assert base_ee.emit("uplifted_event")

    call_me.assert_has_calls(
        [
            call("base event on base emitter"),
            call("shared event on base emitter"),
            call("shared event on uplifted emitter"),
            call("uplifted event on uplifted emitter"),
        ]
    )

    call_me.reset_mock()

    # Quick check for unwrap
    unwrap(uplifted_ee)

    with pytest.raises(AttributeError):
        getattr(uplifted_ee, "unwrap")

    with pytest.raises(AttributeError):
        getattr(base_ee, "unwrap")

    assert not uplifted_ee.emit("base_event")
    assert uplifted_ee.emit("shared_event")
    assert uplifted_ee.emit("uplifted_event")

    assert base_ee.emit("base_event")
    assert base_ee.emit("shared_event")
    assert not base_ee.emit("uplifted_event")

    call_me.assert_has_calls(
        [
            # No listener for base event on uplifted
            call("shared event on uplifted emitter"),
            call("uplifted event on uplifted emitter"),
            call("base event on base emitter"),
            call("shared event on base emitter"),
            # No listener for uplifted event on uplifted
        ]
    )


@pytest.mark.parametrize("error_handling", ["new", "underlying", "neither"])
def test_exception_handling(error_handling):
    base_ee = EventEmitter()
    uplifted_ee = uplift(UpliftedEventEmitter, base_ee, error_handling=error_handling)

    # Exception handling always prefers uplifted
    base_error = Exception("base error")
    uplifted_error = Exception("uplifted error")

    # Hold my beer
    base_error_handler = Mock()
    base_ee._emit_handle_potential_error = base_error_handler

    # Hold my other beer
    uplifted_error_handler = Mock()
    uplifted_ee._emit_handle_potential_error = uplifted_error_handler

    base_ee.emit("error", base_error)
    uplifted_ee.emit("error", uplifted_error)

    if error_handling == "new":
        base_error_handler.assert_not_called()
        uplifted_error_handler.assert_has_calls(
            [call("error", base_error), call("error", uplifted_error)]
        )
    elif error_handling == "underlying":
        base_error_handler.assert_has_calls(
            [call("error", base_error), call("error", uplifted_error)]
        )
        uplifted_error_handler.assert_not_called()
    elif error_handling == "neither":
        base_error_handler.assert_called_once_with("error", base_error)
        uplifted_error_handler.assert_called_once_with("error", uplifted_error)
    else:
        raise Exception("unrecognized setting")


@pytest.mark.parametrize(
    "proxy_new_listener", ["both", "neither", "forward", "backward"]
)
def test_proxy_new_listener(proxy_new_listener):
    call_me = Mock()

    base_ee = EventEmitter()

    uplifted_ee = uplift(
        UpliftedEventEmitter, base_ee, proxy_new_listener=proxy_new_listener
    )

    @base_ee.on("new_listener")
    def base_new_listener_handler(event, f):
        assert event in ("event", "new_listener")
        call_me("base new listener handler", f)

    @uplifted_ee.on("new_listener")
    def uplifted_new_listener_handler(event, f):
        assert event in ("event", "new_listener")
        call_me("uplifted new listener handler", f)

    def fresh_base_handler():
        pass

    def fresh_uplifted_handler():
        pass

    base_ee.on("event", fresh_base_handler)
    uplifted_ee.on("event", fresh_uplifted_handler)

    if proxy_new_listener == "both":
        call_me.assert_has_calls(
            [
                call("base new listener handler", fresh_base_handler),
                call("uplifted new listener handler", fresh_base_handler),
                call("uplifted new listener handler", fresh_uplifted_handler),
                call("base new listener handler", fresh_uplifted_handler),
            ]
        )
    elif proxy_new_listener == "neither":
        call_me.assert_has_calls(
            [
                call("base new listener handler", fresh_base_handler),
                call("uplifted new listener handler", fresh_uplifted_handler),
            ]
        )
    elif proxy_new_listener == "forward":
        call_me.assert_has_calls(
            [
                call("base new listener handler", fresh_base_handler),
                call("uplifted new listener handler", fresh_base_handler),
                call("uplifted new listener handler", fresh_uplifted_handler),
            ]
        )
    elif proxy_new_listener == "backward":
        call_me.assert_has_calls(
            [
                call("base new listener handler", fresh_base_handler),
                call("uplifted new listener handler", fresh_uplifted_handler),
                call("base new listener handler", fresh_uplifted_handler),
            ]
        )
    else:
        raise Exception("unrecognized proxy_new_listener")



# -*- coding: utf-8 -*-

"""
pyee supplies a `EventEmitter` class that is similar to the
`EventEmitter` class from Node.js. In addition, it supplies the subclasses
`AsyncIOEventEmitter`, `TwistedEventEmitter` and `ExecutorEventEmitter`
for supporting async and threaded execution with asyncio, twisted, and
concurrent.futures Executors respectively, as supported by the environment.

# Example

```text
In [1]: from pyee.base import EventEmitter

In [2]: ee = EventEmitter()

In [3]: @ee.on('event')
   ...: def event_handler():
   ...:     print('BANG BANG')
   ...:

In [4]: ee.emit('event')
BANG BANG

In [5]:
```

"""

from pyee.base import EventEmitter, Handler, PyeeError, PyeeException

__all__ = ["EventEmitter", "Handler", "PyeeError", "PyeeException"]



# -*- coding: utf-8 -*-
from unittest.mock import Mock

import pytest

from pyee import EventEmitter
from pyee.cls import evented, on


@evented
class EventedFixture:
    def __init__(self):
        self.call_me = Mock()

    @on("event")
    def event_handler(self, *args, **kwargs):
        self.call_me(self, *args, **kwargs)


_custom_event_emitter = EventEmitter()


@evented
class CustomEmitterFixture:
    def __init__(self):
        self.call_me = Mock()
        self.event_emitter = _custom_event_emitter

    @on("event")
    def event_handler(self, *args, **kwargs):
        self.call_me(self, *args, **kwargs)


class InheritedFixture(EventedFixture):
    pass


@pytest.mark.parametrize(
    "cls", [EventedFixture, CustomEmitterFixture, InheritedFixture]
)
def test_evented_decorator(cls):
    inst = cls()

    inst.event_emitter.emit("event", "emitter is emitted!")

    inst.call_me.assert_called_once_with(inst, "emitter is emitted!")

    _custom_event_emitter.remove_all_listeners()



# -*- coding: utf-8 -*-

import pytest
import pytest_trio.plugin  # type: ignore # noqa
import trio

from pyee.trio import TrioEventEmitter


class PyeeTestError(Exception):
    pass


@pytest.mark.trio
async def test_trio_emit():
    """Test that the trio event emitter can handle wrapping
    coroutines
    """

    async with TrioEventEmitter() as ee:
        should_call = trio.Event()

        @ee.on("event")
        async def event_handler():
            should_call.set()

        ee.emit("event")

        result = False
        with trio.move_on_after(0.1):
            await should_call.wait()
            result = True

        assert result


@pytest.mark.trio
async def test_trio_once_emit():
    """Test that trio event emitters also wrap coroutines when
    using once
    """

    async with TrioEventEmitter() as ee:
        should_call = trio.Event()

        @ee.once("event")
        async def event_handler():
            should_call.set()

        ee.emit("event")

        result = False
        with trio.move_on_after(0.1):
            await should_call.wait()
            result = True

        assert result


@pytest.mark.trio
async def test_trio_error():
    """Test that trio event emitters can handle errors when
    wrapping coroutines
    """

    async with TrioEventEmitter() as ee:
        send, rcv = trio.open_memory_channel(1)

        @ee.on("event")
        async def event_handler():
            raise PyeeTestError()

        @ee.on("error")
        async def handle_error(exc):
            async with send:
                await send.send(exc)

        ee.emit("event")

        result = None
        with trio.move_on_after(0.1):
            async with rcv:
                result = await rcv.__anext__()

        assert isinstance(result, PyeeTestError)


@pytest.mark.trio
async def test_sync_error(event_loop):
    """Test that regular functions have the same error handling as coroutines"""

    async with TrioEventEmitter() as ee:
        send, rcv = trio.open_memory_channel(1)

        @ee.on("event")
        def sync_handler():
            raise PyeeTestError()

        @ee.on("error")
        async def handle_error(exc):
            async with send:
                await send.send(exc)

        ee.emit("event")

        result = None
        with trio.move_on_after(0.1):
            async with rcv:
                result = await rcv.__anext__()

        assert isinstance(result, PyeeTestError)



# -*- coding: utf-8 -*-

from time import sleep
from unittest.mock import Mock

from pyee.executor import ExecutorEventEmitter


class PyeeTestError(Exception):
    pass


def test_executor_emit():
    """Test that ExecutorEventEmitters can emit events."""
    with ExecutorEventEmitter() as ee:
        should_call = Mock()

        @ee.on("event")
        def event_handler():
            should_call(True)

        ee.emit("event")
        sleep(0.1)

        should_call.assert_called_once()


def test_executor_once():
    """Test that ExecutorEventEmitters also emit events for once."""
    with ExecutorEventEmitter() as ee:
        should_call = Mock()

        @ee.once("event")
        def event_handler():
            should_call(True)

        ee.emit("event")
        sleep(0.1)

        should_call.assert_called_once()


def test_executor_error():
    """Test that ExecutorEventEmitters handle errors."""
    with ExecutorEventEmitter() as ee:
        should_call = Mock()

        @ee.on("event")
        def event_handler():
            raise PyeeTestError()

        @ee.on("error")
        def handle_error(e):
            should_call(e)

        ee.emit("event")

        sleep(0.1)

        should_call.assert_called_once()



from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Iterator, List, Type, TypeVar

from pyee import EventEmitter


@dataclass
class Handler:
    event: str
    method: Callable


class Handlers:
    def __init__(self) -> None:
        self._handlers: List[Handler] = []

    def append(self, handler) -> None:
        self._handlers.append(handler)

    def __iter__(self) -> Iterator[Handler]:
        return iter(self._handlers)

    def reset(self):
        self._handlers = []


_handlers = Handlers()


def on(event: str) -> Callable[[Callable], Callable]:
    """
    Register an event handler on an evented class. See the `evented` class
    decorator for a full example.
    """

    def decorator(method: Callable) -> Callable:
        _handlers.append(Handler(event=event, method=method))
        return method

    return decorator


def _bind(self: Any, method: Any) -> Any:
    @wraps(method)
    def bound(*args, **kwargs) -> Any:
        return method(self, *args, **kwargs)

    return bound


Cls = TypeVar("Cls", bound=Type)


def evented(cls: Cls) -> Cls:
    """
    Configure an evented class.

    Evented classes are classes which use an EventEmitter to call instance
    methods during runtime. To achieve this without this helper, you would
    instantiate an `EventEmitter` in the `__init__` method and then call
    `event_emitter.on` for every method on `self`.

    This decorator and the `on` function help make things look a little nicer
    by defining the event handler on the method in the class and then adding
    the `__init__` hook in a wrapper:

    ```py
    from pyee.cls import evented, on

    @evented
    class Evented:
        @on("event")
        def event_handler(self, *args, **kwargs):
            print(self, args, kwargs)

    evented_obj = Evented()

    evented_obj.event_emitter.emit(
        "event", "hello world", numbers=[1, 2, 3]
    )
    ```

    The `__init__` wrapper will create a `self.event_emitter: EventEmitter`
    automatically but you can also define your own event_emitter inside your
    class's unwrapped `__init__` method. For example, to use this
    decorator with a `TwistedEventEmitter`::

    ```py
    @evented
    class Evented:
        def __init__(self):
            self.event_emitter = TwistedEventEmitter()

        @on("event")
        async def event_handler(self, *args, **kwargs):
            await self.some_async_action(*args, **kwargs)
    ```
    """
    handlers: List[Handler] = list(_handlers)
    _handlers.reset()

    og_init: Callable = cls.__init__

    @wraps(cls.__init__)
    def init(self: Any, *args: Any, **kwargs: Any) -> None:
        og_init(self, *args, **kwargs)
        if not hasattr(self, "event_emitter"):
            self.event_emitter = EventEmitter()

        for h in handlers:
            self.event_emitter.on(h.event, _bind(self, h.method))

    cls.__init__ = init

    return cls



# -*- coding: utf-8 -*-

from typing import Any, Generator
from unittest.mock import Mock

from twisted.internet.defer import Deferred, inlineCallbacks, succeed
from twisted.python.failure import Failure

from pyee.twisted import TwistedEventEmitter


class PyeeTestError(Exception):
    pass


def test_emit() -> None:
    """Test that TwistedEventEmitter can handle wrapping
    coroutines
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on("event")
    async def event_handler() -> None:
        _ = await succeed("yes!")
        should_call(True)

    ee.emit("event")

    should_call.assert_called_once()


def test_once() -> None:
    """Test that TwistedEventEmitter also wraps coroutines for
    once
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.once("event")
    async def event_handler():
        _ = await succeed("yes!")
        should_call(True)

    ee.emit("event")

    should_call.assert_called_once()


def test_error() -> None:
    """Test that TwistedEventEmitters handle Failures when wrapping coroutines."""
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on("event")
    async def event_handler():
        raise PyeeTestError()

    @ee.on("failure")
    def handle_error(e):
        should_call(e)

    ee.emit("event")

    should_call.assert_called_once()


def test_propagates_failure():
    """Test that TwistedEventEmitters can propagate failures
    from twisted Deferreds
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on("event")
    @inlineCallbacks
    def event_handler() -> Generator[Deferred[object], object, None]:
        d: Deferred[Any] = Deferred()
        d.callback(Failure(PyeeTestError()))
        yield d

    @ee.on("failure")
    def handle_failure(f: Any) -> None:
        assert isinstance(f, Failure)
        should_call(f)

    ee.emit("event")

    should_call.assert_called_once()


def test_propagates_sync_failure():
    """Test that TwistedEventEmitters can propagate failures
    from twisted Deferreds
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on("event")
    def event_handler():
        raise PyeeTestError()

    @ee.on("failure")
    def handle_failure(f):
        assert isinstance(f, Failure)
        should_call(f)

    ee.emit("event")

    should_call.assert_called_once()


def test_propagates_exception():
    """Test that TwistedEventEmitters propagate failures as exceptions to
    the error event when no failure handler
    """

    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on("event")
    @inlineCallbacks
    def event_handler() -> Generator[Deferred[object], object, None]:
        d: Deferred[Any] = Deferred()
        d.callback(Failure(PyeeTestError()))
        yield d

    @ee.on("error")
    def handle_error(exc):
        assert isinstance(exc, Exception)
        should_call(exc)

    ee.emit("event")

    should_call.assert_called_once()
